/*
 * This file is part of
 *    ivil - Interactive Verification on Intermediate Language
 *
 * Copyright (C) 2009-2012 Karlsruhe Institute of Technology
 *
 * The system is protected by the GNU General Public License.
 * See LICENSE.TXT (distributed with this file) for details.
 */
package de.uka.iti.pseudo.parser.term;

import java.util.Collections;
import java.util.List;

import nonnull.NonNull;
import nonnull.Nullable;
import de.uka.iti.pseudo.parser.ASTVisitException;
import de.uka.iti.pseudo.parser.ASTVisitor;
import de.uka.iti.pseudo.parser.ParserConstants;
import de.uka.iti.pseudo.parser.Token;
import de.uka.iti.pseudo.parser.program.ASTStatement;
import de.uka.iti.pseudo.term.Modality;

/**
 * The AST class for program terms generated by the
 * {@link de.uka.iti.pseudo.parser.Parser}.
 *
 * This class is used for schematic and literal program terms
 */
public class ASTProgramTerm extends ASTTerm {

    private final Token position;
    private @Nullable
    Token programReference;
    private final Modality modality;

    private ASTProgramTerm(Token position, Modality modality) {
        super(Collections.<ASTTerm> emptyList());
        this.modality = modality;
        this.position = position;
        this.programReference = null;
    }

    /**
     * Instantiates a new program term AST.
     *
     * @param label
     *            the fist token
     * @param modality
     *            the modality
     * @param matchStatement
     *            the matching statement
     */
    public ASTProgramTerm(Token label, @Nullable Modality modality,
            ASTStatement matchStatement) {
        this(label, modality);
        if (matchStatement != null) {
            addChild(matchStatement);
        }
    }

    /**
     * Instantiates a new program term AST.
     *
     * @param label
     *            the fist token
     * @param modality
     *            the modality
     * @param programReference
     *            a token reference to the program of the program term
     */
    public ASTProgramTerm(Token label, Modality modality,
            @NonNull Token programReference) {
        this(label, modality);
        this.programReference = programReference;
    }

    @Override
    public Token getLocationToken() {
        return position;
    }

    // Checkstyle: IGNORE JavadocMethod - is a getter
    public Token getProgramReferenceToken() {
        assert !isSchema() : "Schema programs have no such token";
        assert programReference != null : "nullness";
        return programReference;
    }

    public Token getLabel() {
        return position;
    }

    public Modality getModality() {
        return modality;
    }

    @Override
    public void visit(ASTVisitor v) throws ASTVisitException {
        v.visit(this);
    }

    // Checkstyle: IGNORE JavadocMethod - is a getter
    public boolean hasMatchingStatement() {
        return isSchema() && getChildren().size() > 1;
    }

    // Checkstyle: IGNORE JavadocMethod - is a getter
    public ASTStatement getMatchingStatement() {
        assert hasMatchingStatement();
        return (ASTStatement) getChildren().get(0);
    }

    public boolean isSchema() {
        return position.kind == ParserConstants.SCHEMA_IDENTIFIER;
    }

    // Checkstyle: IGNORE JavadocMethod - is a getter
    public ASTTerm getSuffixFormula() {
        List<ASTTerm> children = getSubterms();
        int size = children.size();
        if (size > 0) {
            return children.get(size - 1);
        } else {
            throw new IllegalStateException("There is no suffix formula! (Should not be parsed!)");
        }
    }

}
